signature COLOR_OPT2 = sig
  include PASS
  val dbg : bool ref
  val color_opt : CDocs.cdocs -> CDocs.cdocs
  end

structure ColorOpt2 :> COLOR_OPT2  =
  struct
    structure S = SMLNG
    structure T = CDocs
    structure C = Context
    structure JL = JoinList
    val {dsay,derr,dbg,ddoc} = Debug.mk ()
    val name = "color-opt"
    structure N = Costs
    type costs =  S.color option -> N.cost
    fun zero_costs _ = (N.zero)
    fun sum_costs (x:costs,y:costs) c = N.add(x c,y c)
    fun min_costs (x:costs,y:costs) c = N.min(x c,y c)
    fun mkcosts ctxt v = let
      val c = (C.Get.color ctxt)
    in if c = v then (N.I 0)
       else if Option.isSome c then N.I 7
	    else N.PosInf
    end
    fun ensure_color ctxt c x c' = let
    in if c = c' then x 
       else (case c of
	       NONE => raise (Fail "can't ensure color") 
	     | SOME c => T.Nd(ctxt,S.Color c,JL.one x))
    end
    fun ensure_ctxt ctxt x c = ensure_color ctxt (C.Get.color ctxt) x c


    fun opts [] = [NONE]
      | opts (x::xs) = (SOME x)::(opts xs)
    fun prod res = let
      fun f((x,y),xs) = (x,y)::xs
    in List.foldl (fn (i,acc) =>
		   List.foldl 
		   (fn (j,acc) => f((i,j),acc)) acc res) [] res

    end

    val colors = [S.Red,S.Blue,S.Cyan,S.Magenta,S.Yellow,S.Black,S.White]
    val colors_opts      = opts colors
    val colors_opts_prod = prod colors_opts

    fun c2s NONE =  "NONE"
      | c2s (SOME v) = (S.tag2str (S.Color v))

    fun min lst = let
      fun loop [] = raise (Fail "no costs")
	| loop [x] = x
	| loop ((x,vx)::rest) = let
	    val (y,vy) = loop rest
	  in if N.lt(x,y) then (x,vx) else (y,vy)
	  end
    in loop lst
    end

    fun min_cost_color costs  = let 
      fun f c = let val cost = costs c
	        in  dsay ("\n# color is "^(c2s c));
		    dsay (" cost is "^(N.toString cost));
		  (cost,c)
                end
    in min (List.map f colors_opts)
    end

    fun min_cost_color_pair (x,y) = let 
      fun f (c1,c2) = (N.add(x c1,y c2),(c1,c2))
    in min (List.map f colors_opts_prod)
    end

	
    fun color_opt doc = let
      fun kNil () = ((fn _ => T.Txt(C.init,"")),zero_costs)
      fun kCons (_,_) = raise (Fail "hole")
(*      fun kSeq((x,cx),(y,cy)) = let
	val (child_cost,(lcolor,rcolor)) = min_cost_color_pair (cx,cy)
	fun costs c = 
	  if lcolor = rcolor then
	    (if c = lcolor then child_cost
	     else (N.add(child_cost,N.I 7)))
	  else if not (lcolor = c) andalso not (rcolor = c)
		 then N.add(child_cost,N.I 14)
	       else  N.add(child_cost,N.I 7)
		 
	fun mk c =
	  if lcolor = rcolor then
	    ensure_color lcolor (S.Seq(x lcolor,y rcolor)) c
	  else
	    S.Seq(ensure_color lcolor (x lcolor) c,   
		  ensure_color rcolor (y rcolor) c)
	val dumb =  ((fn v => S.Seq(x v,y v)),sum_costs(cx,cy))
      in (mk,costs)
      end
*)
      fun kTxt(c,"") = ((fn _ => T.Txt(C.init,"")),zero_costs)
        | kTxt(c,s) = (ensure_ctxt c (T.Txt(c,s)),mkcosts c)

      fun kWs(c,s)  =
	if C.Get.u c = 0 then (fn _ => (S.Ws s),zero_costs)
	else (ensure_ctxt c (S.Ws s),mkcosts c)

      fun kNd(c,S.Color color,(d,cd)) = let
	fun costs (SOME v) = let
	  val child_cost = cd (SOME v)
	in if (v = color) then N.sub(child_cost,N.I 7)
	   else child_cost
	end
	  | costs NONE = (cd NONE)
	fun mk (SOME v) =
	  if (v = color) then d (SOME v)
	  else S.Nd(S.Color color,d (SOME color))
	  | mk NONE = d NONE
      in (mk,costs)
      end 
      | kNd(c,t,(d,cd)) = ((fn c => (S.Nd(t,d c))),cd)
      val (mk,costs) =
	C.fold_with_ctxt {kSeq=kSeq, kTxt=kTxt, kWs=kWs, kNd=kNd} 
	C.init doc

      fun top_costs NONE = costs NONE
	| top_costs(SOME v) = N.add(costs (SOME v),N.I 7) 

      val (cost,mincolor) = min_cost_color top_costs
      val _ = dsay ("\n# top color is "^(c2s mincolor))
      val _ = dsay (" cost is "^(N.toString cost))


    in if N.lt(cost,N.zero) then ensure_color mincolor (mk mincolor) NONE
       else (dsay "\n# Can't improve";doc)
    end

    val tests = Tests.mk name false go
      ["a",
       "<r>a</r>",
       "<r></r><b></b>",
       "<r></r><b></b><r></r><g><r></r><b></b><r></r></g>",
       "<r>a</r><r>a</r>",
       "<b>a</r><r>a</r>"
       ]

  end

