signature CONTEXT = sig
  type ctxt

  structure Get : sig
    val b       : ctxt -> bool
    val em      : ctxt -> bool
    val i       : ctxt -> bool
    val s       : ctxt -> bool
    val tt      : ctxt -> bool
    val u       : ctxt -> word
    val size    : ctxt -> word option
    val color   : ctxt -> SMLNG.color option
  end

  val init   : ctxt
  val upd    : SMLNG.tag * ctxt -> ctxt
  val eq     : (ctxt * ctxt) -> bool

  (* compute the space context *)
  (* b, em, i, s : true
     u : 0 > 3
     size: NONE
     color : if u = 0 then x else SOME white
  *)
  val upd_ws : ctxt -> ctxt

  (* sets flag associated with tag to initial default *)
  (* b,em,i,s,tt : false
     u : 0
     size, color : NONE
     pl : reset all pl fields
   *)
  val reset : SMLNG.tag * ctxt -> ctxt
  val reset_color : ctxt -> ctxt
  val reset_size  : ctxt -> ctxt

  (* check if tag changes context *)
  (* b, i, s, tt : x
     em : if s then true else false
     pl : s = em = i = b = tt = false andalso u = 0
     u : x >= 3
     size, color : t = x
  *)
  val isNop  : SMLNG.tag * ctxt -> bool

  (* compute intersection *)
  (* b, em, i, s, tt: x andalso y
     u : min(u1,u2)
     size, color : if x = y then x else NONE
  *)
  val intersect : (ctxt * ctxt) -> ctxt

  (* compute difference *)
  (* b, em, i, s, tt : x andalso (not y)
     u : if u1 > u2 then u1 - u2 else 0
     size, color : if x = y then NONE else x
  *)
  val difference : (ctxt * ctxt) -> ctxt

  (* append a context to another context;x is appended to y*)
  (* b, i, tt, s : x orelse y
     em : if s_x orelse s_y then false else em_x xor em_y
     u : u1+u2
     size, color : if x=NONE then y else x
  *)
  val append: ctxt * ctxt ->ctxt

  val fold_with_ctxt : {kSeq:('a * 'a) -> 'a,
			kTxt:(ctxt * string) -> 'a,
			kWs:(ctxt * string) -> 'a,
			kNd:(ctxt * SMLNG.tag * 'a) -> 'a} ->
                                     ctxt -> SMLNG.doc -> 'a

  val toTags : ctxt -> SMLNG.tag list
  end
