signature COSTS = sig
 datatype cost = PosInf | NegInf | I of int

 val zero : cost
 val add  : cost * cost -> cost
 val sub  : cost * cost -> cost
 val max  : cost * cost -> cost
 val min  : cost * cost -> cost
   
 val compare : cost * cost -> order
 val lt : cost * cost -> bool
 val gt : cost * cost -> bool
 val eq : cost * cost -> bool
 val toString : cost -> string
   
end
structure Costs :> COSTS =
    struct
      datatype cost = PosInf | NegInf | I of int
      val zero = I 0
      fun add (I x,I y) = (I (x + y))
	| add (PosInf,NegInf) = raise (Fail "undefined")
	| add (NegInf,PosInf) = raise (Fail "undefined")
	| add (PosInf, _) = PosInf
	| add (NegInf, _) = NegInf
	| add (_,PosInf)  = PosInf
	| add (_,NegInf)  = NegInf

      fun sub (I x,I y) = (I (x - y))
	| sub (PosInf,NegInf) = raise (Fail "undefined")
	| sub (NegInf,PosInf) = raise (Fail "undefined")
	| sub (PosInf, _) = PosInf
	| sub (NegInf, _) = NegInf
	| sub (_,PosInf)  = PosInf
	| sub (_,NegInf)  = NegInf

      fun min (I x,I y) = (I (Int.min(x,y)))
	| min (x,NegInf) = NegInf
	| min (NegInf,x) = NegInf
	| min (PosInf,x) = x
	| min (x,PosInf) = x


      fun max (I x,I y) = (I (Int.max(x,y)))
	| max (x,NegInf) = x
	| max (NegInf,x) = x
	| max (PosInf,x) = PosInf
	| max (x,PosInf) = PosInf

      fun compare(I x,I y) = Int.compare(x,y)
	| compare(NegInf,NegInf) = EQUAL
	| compare(I _,NegInf) = GREATER
	| compare(NegInf,I _) = LESS
	| compare(PosInf,PosInf) = EQUAL
	| compare(PosInf,I _) = GREATER
	| compare(I _,PosInf) = LESS
	| compare (_,_) = raise (Fail "undefined")
      fun lt (x,y) = compare(x,y) = LESS
      fun gt (x,y) = compare(x,y) = GREATER
      fun eq (x,y) = compare(x,y) = EQUAL
      fun toString PosInf = "+Inf"
	| toString NegInf = "-Inf"
	| toString (I c)  = Int.toString c
end