signature DEBUG = sig
  (* turn off all debugging output *)
  val quiet : unit -> unit
  (* turn on global debug flag *)
  val noisy : unit -> unit
  val mk : unit -> {dbg : bool ref,
		    derr:string -> exn,
		    dsay:string -> unit,
		    ddoc:SMLNG.doc -> unit}
  end
structure Debug :> DEBUG =
  struct
    val glbl_dbg = ref true
    fun quiet () = glbl_dbg := false
    fun noisy () = glbl_dbg := true
    fun mk () = let
      val dbg = ref true
      fun dsay x = if ((!glbl_dbg) andalso (!dbg)) then
	TextIO.output(TextIO.stdErr,x) else ()
(*      fun dsay x = ()*)
      fun derr s = let
      in dsay s;Fail s
      end
      fun ddoc doc =
	if ((!glbl_dbg) andalso (!dbg))
	  then SMLNG.print doc TextIO.stdErr else ()
    in {dbg=dbg,dsay=dsay,derr=derr,ddoc=ddoc}
    end
  end