signature MERGE_ADJACENT = sig
  include PASS
  val dbg : bool ref
  val merge_adj : CDocs.cdocs -> CDocs.cdocs
end
structure MergeAdjacent :> MERGE_ADJACENT  =
  struct
    structure T = CDocs
    structure C = Context
    val {dsay,derr,ddoc,dbg} = Debug.mk ()
    structure JL = JoinList
    fun mkDocs c (NONE,grp,d) = JL.append(grp,d)
      | mkDocs c (SOME t,grp,d) = let
	in JL.append(JL.one(T.Nd(c,t,grp)),d)
	end

    fun merge_adj cdocs = let
      val kTxt = T.Txt
      val kWs = T.Ws
      fun kNd(c,t,mi) = T.Nd(c,t,mkDocs (C.upd(t,c)) mi)
      fun kNil () = (NONE,JL.empty,JL.empty)
      fun kCons(T.Nd(c,t,ds),(NONE,grp,rest)) =
	(SOME t,ds,mkDocs c (NONE,grp,rest))
	| kCons(T.Nd(c,t,ds),(SOME t',grp,rest)) =
	if t = t' then (SOME t,JL.append(ds,grp),rest)
	else (SOME t,ds,mkDocs c (SOME t',grp,rest))
	| kCons (c,(x,grp,rest)) =
	  (NONE,JL.one c,mkDocs (T.ctxt c) (x,grp,rest))  
      val res =
	T.fold {kNil=kNil, kCons=kCons, kTxt=kTxt, kWs=kWs, kNd=kNd} cdocs
    in (mkDocs C.init res)
    end

    val name = "merge-adj"
    val go = T.toCleanDoc o merge_adj o T.fromDoc
    val tests = Tests.mk name false go 
      ["1    ",
       "<b></b><r></r> ",
       "1 <b> 2 </b> 3 ",
       "<b>1</b><b>2</b> ",
       "<b>1<b>2</b></b> ",
       "<b>1</b><r>2</r> ",
       "<b>1<b>2</b></b><b>3</b>a",
       "1<b>2</b></b><b>3",
       "<b>1<r>2</r><r>3</r></b><b>4</b>"]

  end
