signature SIZE_OPT = sig
  include PASS
  val dbg : bool ref
  end

structure SizeOpt :> SIZE_OPT =
  struct
    structure T = SMLNG
    structure C = Context
    val {dsay,derr,dbg,ddoc} = Debug.mk ()
    val name = "size-opt"
    structure N = Costs
    type costs =  word option -> N.cost
    fun zero_costs _ = (N.zero)
    fun sum_costs (x:costs,y:costs) c = N.add(x c,y c)
    fun min_costs (x:costs,y:costs) c = N.min(x c,y c)
    fun mkcosts ctxt v = let
      val c = (C.Get.size ctxt)
    in if c = v then (N.I 0)
       else if Option.isSome c then N.I 7
	    else N.PosInf
    end
    fun ensure_size c x c' = let
    in if c = c' then x 
       else (case c of
	       NONE => raise (Fail "can't ensure size") 
	     | SOME c => T.Nd(T.Size c,x))
    end
    fun ensure_ctxt ctxt x c = ensure_size (C.Get.size ctxt) x c


    fun opts [] = [NONE]
      | opts (x::xs) = (SOME x)::(opts xs)
    fun prod res = let
      fun f((x,y),xs) = (x,y)::xs
    in List.foldl (fn (i,acc) =>
		   List.foldl 
		   (fn (j,acc) => f((i,j),acc)) acc res) [] res

    end

    val sizes = [0w0,0w1,0w2,0w3,0w4,0w5,0w6,0w7,0w8,0w9]
    val sizes_opts      = opts sizes
    val sizes_opts_prod = prod sizes_opts

    fun c2s NONE =  "NONE"
      | c2s (SOME v) = (T.tag2str (T.Size v))

    fun min lst = let
      fun loop [] = raise (Fail "no costs")
	| loop [x] = x
	| loop ((x,vx)::rest) = let
	    val (y,vy) = loop rest
	  in if N.lt(x,y) then (x,vx) else (y,vy)
	  end
    in loop lst
    end

    fun min_cost_size costs  = let 
      fun f c = let val cost = costs c
	        in  dsay ("\n# size is "^(c2s c));
		    dsay (" cost is "^(N.toString cost));
		  (cost,c)
                end
    in min (List.map f sizes_opts)
    end
    fun min_cost_size_pair (x,y) = let 
      fun f (c1,c2) = (N.add(x c1,y c2),(c1,c2))
    in min (List.map f sizes_opts_prod)
    end

    fun go doc = let
      fun kSeq((x,cx),(y,cy)) = let
	val (child_cost,(lsize,rsize)) = min_cost_size_pair (cx,cy)
	fun costs c = 
	  if lsize = rsize then
	    (if c = lsize then child_cost
	     else (N.add(child_cost,N.I 7)))
	  else if not (lsize = c) andalso not (rsize = c)
		 then N.add(child_cost,N.I 14)
	       else  N.add(child_cost,N.I 7)
		 
	fun mk c =
	  if lsize = rsize then
	    ensure_size lsize (T.Seq(x lsize,y rsize)) c
	  else
	    T.Seq(ensure_size lsize (x lsize) c,   
		  ensure_size rsize (y rsize) c)
	val dumb =  ((fn v => T.Seq(x v,y v)),sum_costs(cx,cy))
      in (mk,costs)
      end
      fun kTxt(c,"") = ((fn _ => T.Txt ""),zero_costs)
	| kTxt(c,s) = (ensure_ctxt c (T.Txt s),mkcosts c)

      fun kWs(c,s)  =
	if C.Get.u c = 0w0 then (fn _ => (T.Ws s),zero_costs)
	else (ensure_ctxt c (T.Ws s),mkcosts c)

      fun kNd(c,T.Size size,(d,cd)) = let
	fun costs (SOME v) = let
	  val child_cost = cd (SOME v)
	in if (v = size) then N.sub(child_cost,N.I 7)
	   else child_cost
	end
	  | costs NONE = (cd NONE)
	
	fun mk (SOME v) = 
	  if  (v = size) then d (SOME v)
	  else T.Nd(T.Size size,d (SOME size))
	  | mk NONE = d NONE
      in (mk,costs)
      end
      | kNd(c,t,(d,cd)) = ((fn c => (T.Nd(t,d c))),cd)
      val (mk,costs) =
	C.fold_with_ctxt {kSeq=kSeq, kTxt=kTxt, kWs=kWs, kNd=kNd} 
	C.init doc

      fun top_costs NONE = costs NONE
	| top_costs(SOME v) = N.add(costs (SOME v),N.I 7) 

      val (cost,minsize) = min_cost_size top_costs
      val _ = dsay ("\n# top size is "^(c2s minsize))
      val _ = dsay (" cost is "^(N.toString cost))


    in if N.lt(cost,N.zero) then ensure_size minsize (mk minsize) NONE
       else (dsay "\n# Can't improve";doc)
    end

    val tests = Tests.mk name false go [""]

  end

