signature STATE_MONAD =  sig
  type ('a,'state) ST
  val ret    : 'a -> ('a,'state) ST
  val bnd    : ('a,'state) ST -> ('a -> ('b,'state) ST)  -> ('b,'state) ST

  val updST  : ('state -> 'state) -> ('a,'state) ST -> ('a,'state) ST
  val bndST  : ('a,'state) ST -> (('a * 'state) -> ('b,'state) ST) ->
              ('b,'state) ST

  val evalST : ('a,'state) ST -> 'state -> ('a * 'state)
  end
structure StateMonad :> STATE_MONAD =
  struct
    datatype ('a,'b) ST = ST of 'b -> ('a * 'b)
    fun eval (ST m) s = m s
    fun ret v = ST (fn s => (v,s))

    fun bnd m f = ST (fn s => let
		     val (v,s) = eval m s
		     in eval (f v) s
		     end)

    (* operations on the state *)
    fun bndST m f = ST (fn s => let
		       val (v,s) = eval m s
		       in eval (f (v,s)) s
		      end)

    fun updST f m = ST (fn s => let
		      val (v,s) = eval m s
		      in (v,f s)
		      end)
    val evalST = eval
  end