signature TESTS = sig
   val dbg : bool ref
   val readTest : string -> string
   val mk :
     string -> bool -> (SMLNG.doc -> SMLNG.doc) -> string list -> 
     (unit -> unit)
   val dflt : string list
  end
structure Tests :> TESTS =
  struct
    val {dbg,dsay,derr,ddoc} = Debug.mk ()
    fun check ignore_size n f old = let
      val new = f old
      val (new_sz,old_sz,same) = Canon.compareDoc (new,old)
      val bigger = new_sz > old_sz
      val failure =
	if not same then (SOME "\n# Not Same")
        else if (not ignore_size) andalso bigger then (SOME "\n# Bigger")
        else NONE
      fun dmsg () =
	case failure of
	  NONE =>  (dsay ("\n# change: "^(Int.toString (old_sz - new_sz))))
	| SOME s =>(dsay s)
    in (new,dmsg,Option.isSome failure)
    end

    fun dfail () =
      (dsay ("\n#########################################################");
       dsay ("\n#                    TEST FAILURE                       #");
       dsay ("\n#########################################################"))

    fun test_str ignore_size n f (in_str) = let
      val doc = SMLNG.parse (TextIO.openString in_str)
      val _ = dsay ("\n###########")
      val _ = dsay ("\n#   test: "^n);
      val _ = dsay ("\n# source: "^in_str);
      val (new,dmsg,fail) = check ignore_size n f doc
    in if fail then dfail() else ();
      dsay ("\n#  input: "); ddoc doc; 
      dsay ("\n# output: "); ddoc new; dmsg(); dsay("\n#")
    end

    fun do_test (n,f,{ignore_size,tests}) = let
      fun loop (i,[]) = dsay "\n##### Done\n"
	| loop (i,x::xs) = 
	((test_str ignore_size (n^" "^(Int.toString i)) f x) ;loop(i+1,xs))
    in dsay ("\n#### Start "^n);loop(1,tests)
    end
  
    fun readTest x = let
      val ins = TextIO.openIn ("../tests/"^x)
      val s = TextIO.inputAll ins
    in TextIO.closeIn ins;s
    end

    fun mk n b f tests =
      (fn () => do_test(n,f,{ignore_size=b,tests=tests}))

    val dflt = 
      ["1",
       "<B>1</B>",
       "<EM>1</EM>",
       "<I>1</I>",
       "<PL>1</PL>",
       "<S>1</S>",
       "<TT> 1 </TT>",
       "<U> 1 </U>",
       "<0> 1 </0>",
       "<b> 1 </b>",
       "1 2",
       "<B>1</B>2",
       "<EM>1</EM>2",
       "<I>1</I>2",
       "<PL>1</PL>2",
       "<S>1</S>2",
       "<TT> 1 </TT>2",
       "<U> 1 </U>2",
       "<0> 1 </0>2",
       "<b> 1 </b>2",
       "<B>1 2</B>",
       "<B><B>1</B>2</B>",
       "<B><EM>1</EM>2</B>",
       "<B><I>1</I></B>2",
       "<B><PL>1</PL>2</B>",
       "<B><S>1</S>2</B>",
       "<B><TT> 1 </TT>2</B>",
       "<B><U> 1 </U>2</B>",
       "<B><0> 1 </0>2</B>",
       "<B><b> 1 </b>2</B>",
       "<EM><B>1 2</B></EM>",
       "<EM><B><B>1</B>2</B></EM>",
       "<EM><B><EM>1</EM>2</B></EM>",
       "<EM><B><I>1</I></B>2</EM>",
       "<EM><B><PL>1</PL>2</B></EM>",
       "<EM><B><S>1</S>2</B></EM>",
       "<EM><B><TT> 1 </TT>2</B></EM>",
       "<EM><B><U> 1 </U>2</B></EM>",
       "<EM><B><0> 1 </0>2</B></EM>",
       "<EM><B><b> 1 </b>2</B></EM>",
       "<EM><PL><B>1 2</B></PL></EM>",
       "<EM><PL><B><B>1</B>2</PL></B></EM>",
       "<EM><B><EM>1</EM>2</B></EM>",
       "<EM><B><I>1</I></B>2</EM>",
       "<EM><B><PL>1</PL>2</B></EM>",
       "<EM><B><S>1</S>2</B></EM>",
       "<EM><B><TT> 1 </TT>2</B></EM>",
       "<EM><B><U> 1 </U>2</B></EM>",
       "<EM><B><0> 1 </0>2</B></EM>",
       "<EM><B><b> 1 </b>2</B></EM>",
       "<r>aaa</r><g>bbb</g><b>ccc</b><g>ddd</g><r>eee</r>",
       "<r>aaa</r></EM><g>bbb</g><PL><b>ccc</b><PL><g>ddd</g></EM><r>eee</r>"]

    fun mk n b f [] =
      (dsay ("# using default tests for "^n^"\n");
      (fn () => do_test(n,f,{ignore_size=b,tests=dflt})))
      | mk n b f tests =
      (fn () => do_test(n,f,{ignore_size=b,tests=tests}))
  end