signature WS_SIMPLIFICATION = sig
  include PASS
end
structure WSSimplification :> WS_SIMPLIFICATION =
  struct
    val {dsay,derr,ddoc,dbg} = Debug.mk ()

    structure T = SMLNG
    structure C = Context
    val name = "ws-simplification"
    fun go doc = let
      val _   = dsay "\n#ws-simpl:\n"
      val _   = ddoc doc

      fun isWs #" " = true
        | isWs _ = false

      fun isColor (T.Color T.Blue) = true
        | isColor (T.Color T.Green) = true
        | isColor (T.Color T.Red) = true
        | isColor (T.Color T.Cyan) = true
        | isColor (T.Color T.Magenta) = true
        | isColor (T.Color T.Yellow) = true
        | isColor (T.Color T.Black) = true
        | isColor (T.Color T.White) = true
        | isColor _ = false

      fun isNewSpaceCtxt (T.S,_) = false
        | isNewSpaceCtxt (T.EM,_) = false
        | isNewSpaceCtxt (T.I,_) = false
        | isNewSpaceCtxt (T.B,_) = false
        | isNewSpaceCtxt (t,u) = if (u andalso (isColor t)) then true
                                                             else false 

      fun has_lead_ws s = isWs(String.sub(s, 0))
      fun has_trail_ws s = isWs(String.sub(s, ((String.size s)-1)))
      fun has_tail s = if (isWs(String.sub(s, ((String.size s)-1)))) then false else true

      fun strip_ws (s,b) =
        let
           fun strip_trail s = (String.substring(s,0,((String.size s)-1)))
           fun strip_lead s = (String.substring(s,1,((String.size s)-1)))
        in 
          if ((String.size s) > 0) then 
          if (has_lead_ws s) then
            (strip_lead s)
          else
           s
          else 
           s 
        end


      fun simp_ws ((T.Ws s), false, u, false)      = ((T.Ws s), true, u, false)
      |   simp_ws ((T.Ws s), false, u, true)       = ((T.Ws s), false, u, true)
      |   simp_ws ((T.Ws s), true, u, false)          = ((T.Ws ""), true, u, false)
      |   simp_ws ((T.Ws s), true, u, true)          = ((T.Ws s), true, u, true) 
      |   simp_ws ((T.Txt " "), true, u, false)           = ((T.Txt ""), true, u, false)
      |   simp_ws ((T.Txt " "), false, u, false)   = ((T.Txt " "), true, u, false)
      |   simp_ws ((T.Txt " "), false, u, true)    = ((T.Txt " "), false, u, true)
      |   simp_ws ((T.Txt " "), true, u, true)          = ((T.Txt " "), true, u, true) 
      |   simp_ws ((T.Txt s), b, u, true)          = ((T.Txt s), b, u, true)
      |   simp_ws ((T.Txt s), true, u, false)      = 
            if (has_trail_ws s) then
               (T.Txt (strip_ws (s,true)), true, u, false)
            else 
               (T.Txt (strip_ws (s,true)), false, u, false)
      |   simp_ws ((T.Txt s), false, u, false)     = 
            if ((String.size s) > 0) then
              if (has_trail_ws s) then
                 (T.Txt s, true, u, false)
              else
                 (T.Txt s, false, u, false)
            else (T.Txt s, false, u, false)
      |   simp_ws (T.Nd(T.PL,d), b, u, tt)      = 
          let
             val (d',b',u',tt')= simp_ws(d, b, u, false) 
          in
             (T.Nd(T.PL,d'), b', u', tt')
          end
      |   simp_ws (T.Nd(T.TT,d), b, u, tt)      = 
          let
             val (d',b',u',tt')= simp_ws(d, b, u, true) 
          in
             (T.Nd(T.TT,d'), b', u', tt')
          end
      |   simp_ws (T.Nd(T.U,d), b, u, tt)    =
          let
             val (d',b',u',tt')= simp_ws(d, false, true, tt) 
          in
             (T.Nd(T.U, d'), b', u', tt')
          end
 
      |   simp_ws (T.Nd(t,d), b, u, tt)         = 
          let
             val (d',b', u',tt')= if (isNewSpaceCtxt (t,u)) then simp_ws(d, false, u, tt) 
                                                    else simp_ws(d, b, u, tt)
          in
              (T.Nd(t, d'), b', u', tt')
          end
      |   simp_ws (T.Seq(x,y), b, u, tt)        = 
          let
             val (x', b', u', tt')=  simp_ws(x, b, u, tt)
             val (y', b'', u'', tt'')= simp_ws(y, b', u', tt')
          in
             (T.Seq(x',y'),b'', u'', tt'')
          end
      
      val (doc', _, _,_) = simp_ws (doc, false, false, false)
 
      val _ = dsay "\nws-simpl:\n"
      val _ = ddoc doc'
    in doc'
    end

    val tests = Tests.mk name false go ["<b> </b>",
       "<B><I><r>bla bla </r> </I> </B> <r><I> barfoo</I></r>",
       "<U><r> xxx </r><b> </b><r> yyy </r></U>",
       "<r> xxx </r><b> </b><r> yyy </r>",
       "abcde <7><PL><B>abcde</B></PL></7> <5>xyz</5>",
       "<TT> <I>a <B> b </B> </I> </TT>",
       "<TT> <I><b>a <B> b </B> </b></I> </TT>",
       "<I>a <B> b</B> </I> ",
       "a <TT><PL> b</PL></TT>",
       "<EM>abcd <TT>a<b><U> ef </U><PL>xyz </PL></TT>ghi </EM>",
       "<EM>abcd <TT>a<b><U> ef </U><PL>xyz <I><B> a</B> </I> </PL></b>ghi </TT></EM>",
       "<TT>ab <PL> cd <TT> ef</TT> </PL> </TT>"
    ]
  end
